// Program iterface for TI-83/TI-86 Small C
//
// Program will not run in a compiler!!!!!!
//
// Tested to work with only the Borland C++ v3.1 & v4.5 compilers

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <io.h>
#include <stdarg.h>
#include <process.h>
#include <dos.h>
#include <dir.h>
#include <ctype.h>

// Programs
#define COMPILER        "8XCC.EXE"
#define PREPROCESSOR 	"DCPP.EXE"
#define OPTIMIZER    	"COPT.EXE"
#define ASSEMBLER    	"TASM.EXE"
#define OBJ86	     	"PRGM86.EXE"
#define OBJ83	     	"DEVPAC83.COM"

// Program file mode types
#define INPUT	 	0x01
#define OUTPUT_INTO	0x02
#define OUTPUT_SCRATCH	0x04

// File mode numbers
#define STDIN		0
#define STDOUT 		1

// Data types
#define DATA_INVALID	0
#define DATA_NUMBER	1
#define DATA_STRING	2
#define DATA_BOOLTRUE	3
#define DATA_BOOLFALSE	4

// Bool
#define TRUE		1
#define FALSE		0

typedef struct tag_option {
	char shortname  [10];		// Shorter name for command line usally
	char longname   [20];		// Longer name for INI file usally
	char datatype;			// Data type
					// 1 = Number
					// 2 = String
					// 3 = BOOL TRUE
	char bSet;			// Set
	void *ptr;			// Pointer to data
} option;

typedef struct tag_filelist {
	char   *fileptr;		// pointer to argv
	struct tag_filelist *next;	// next item
} filelist;

filelist *FileList = NULL;
filelist *LastFileList = NULL;

// Data types
char szIncludePath[128]    = "INCLUDE";
char szLibaryPath[128]     = "INCLUDE";
char szAssemblyPath[128]   = "ASM";
char szExecutablePath[128] = "EXE";
char szIniFile[128]	   = "GO8X.INI";
char szOldDir[MAXDIR]  = "";
int nologo = FALSE;
int noopt = FALSE;
int compileonly = FALSE;
int keepinter = FALSE;
int nCalcMode = -1;

#define GetMemory(size) _GetMemory(size, __FILE__, __LINE__)

option optionlist[] = {
  { "noopt",   "",           3, FALSE, &noopt           },
  { "compile", "",           3, FALSE, &compileonly     },
  { "keep",    "",           3, FALSE, &keepinter       },
  { "nologo",  "",           3, FALSE, &nologo          },
  { "I",       "INCLUDE",    2, FALSE, szIncludePath    },
  { "L",       "LIBARY",     2, FALSE, szLibaryPath     },
  { "A",       "ASSEMBLER",  2, FALSE, szAssemblyPath   },
  { "E",       "EXECUTABLE", 2, FALSE, szExecutablePath },
  { "i",       "",           2, FALSE, szIniFile        },
  { "c",       "",           1, FALSE, &nCalcMode       },
};

const int optionlistsize = (sizeof(optionlist)) / sizeof(struct tag_option);

void RestoreDirectory(void);

#define terminate(x) { RestoreDirectory(); exit(x); }

void help(void)
{
	printf("\
GO8X -cxx [options] file\n\
\n\
  -c is the calculator mode.  Valid values are 83 for TI-83 and 86 for TI-86\n\
\n\
	Options:\n\
\n\
	-Ipath  = Include path		       -Lpath  = Libary path\n\
	-Epath  = Executable Files path	       -Apath  = Asseblyer path\n\
	-nologo = Do not print logo	       -compile= Compile to .ASM only\n\
	-noopt  = Do not optimize	       -keep   = Keep Intermedate Files\n\
	-ifile  = INI File name (default: GO8X.INI)\n\
");
    terminate(0);
}

int exist(char *file) {
  FILE *testinger;
  if((testinger = fopen(file,"r"))==NULL)
    return 0;
  fclose(testinger);
  return 1;
  }

char *move_up(char *mvs)
{
	return (mvs++);
}

int atoi2(char s[], int *err) {
	int i,n,sign;
	*err = 0;
	i=0;

	sign = 1;
	switch(s[i]){
	case '-': sign = -1; /* and fall through */
	case '+': ++i;
		break;
	}
	n = 0;
	while(isdigit(s[i]))
	{
		n = 10 * n + s[i] - '0';
		i++;
	}
	if(s[i] != 0x00 && s[i] != '\r' && s[i] != '\n')
	{
		*err = 1;
		return 0;
	}
	return (sign * n);
}

int pro_spawn_to_file(char *infile, char *outfile, char writeselect,
		char *name, ...) {
	int oldstdout, oldstdin;
	FILE *stdoutbackup;
	unsigned char errcode;

	// Copy modes
	oldstdin  = dup(STDIN);
	oldstdout = dup(STDOUT);

	// Check file modes
	if((writeselect & OUTPUT_INTO) && (writeselect & OUTPUT_SCRATCH))
	{
		printf("Codding error invalid writingselect");
        terminate(1);
	}

	// Do file moding
	if(writeselect & INPUT)
	{
		if((freopen(infile, "r",stdin))==NULL) {
			printf("Could not open file %s\n", infile);
            terminate(3);
		}
	}
	if(writeselect & OUTPUT_INTO)
	{
		if((stdoutbackup = freopen(outfile,"a",stdout))==NULL) {
			printf("Could not write file %s\n",outfile);
	    terminate(1);
		}
	fseek(stdoutbackup, 0L, SEEK_END);
	}
	if(writeselect & OUTPUT_SCRATCH)
	{
		if((stdoutbackup = freopen(outfile, "w", stdout))==NULL) {
			printf("Could not write file %s\n",outfile);
            terminate(2);
		}
	}
	fprintf(stdoutbackup, "");
	va_list argptr;

	va_start(argptr,name);

	// first type the three types
	errcode = spawnvp(P_WAIT, name, (char **)argptr);
	va_end(argptr);
	dup2(oldstdout,STDOUT);
	dup2(oldstdin,STDIN);
	close(oldstdout);
	close(oldstdin);

    if(errcode != 0)
        terminate(errcode);          // Pass this errcode to the end of program

	return errcode;
}

char *catpath(char *path, char *add)
{
	if(path[strlen(path)-1] != '\\' && add[0] != '\\')
		strcat(path, "\\");
	strcat(path, add);
	return path;
}

char *setuppath(char *exename)
{
	char drive[MAXDRIVE];
	char dir[MAXDIR];
	char file[MAXFILE];
	char ext[MAXEXT];
	char full[2048];
	int  flags;

	flags = fnsplit (exename,drive,dir,file,ext);

	if(!(flags & DRIVE))
		sprintf(drive,"%c:\\", getdisk() + 'A');
	if(!(flags & DIRECTORY))
		getcurdir(drive[0]-'A',dir);

	strcpy (full,drive);
	catpath(full,dir);

	return full;
}

char *trim(char *oldtext)
  {
  int x;
  char newbe[1024];
  strcpy(newbe,oldtext);
  while(newbe[0] == ' ')
     strcpy(newbe, (newbe+1));

  x = 0;
  while(newbe[x] != 0x00)
  {
    if(newbe[x] == ';')
    {
      newbe[x] = 0x00;
      break;
    }
    x++;
  }

  while(newbe[strlen(newbe)-1] == '\n' || newbe[strlen(newbe)-1] == '\r')
    newbe[strlen(newbe)-1] = 0x00;
  while(newbe[strlen(newbe)-1] == ' ')
    newbe[strlen(newbe)-1] = 0x00;
  return newbe;
  }

void ProcessINIFile(char *szFileName)
{
	char *option, *value;
	char line[1024];
	char line2[1024];
	int cCountOption, bFoundIt;
	FILE *pConfigFile;
	if((pConfigFile = fopen(szFileName, "rt")) == NULL)
	{
		printf("Error reading %s configuation file\n", szFileName);
        terminate(1);
	}
	while(fgets(line,1024,pConfigFile) != NULL)
	{
		strcpy(line2, trim(line));
		if(line2[0] == 0x00)
			continue;	// Line blank or has a comment only

		option = line2;
		value = option;
		while(value[0] != '=' && value[0] != 0x00)
			value++;
		if(value[0] == 0x00)
		{
			printf("Error: expected = in line:\n%s\n",line);
            terminate(5);
		}
		value[0] = 0x00;
		value++;
		strcpy(option, trim(option));
		strcpy(value, trim(value));
		if(option[0] == 0x00)
		{
			printf("Error: What do I set in line:\n\%s\n", line);
            terminate(5);
		}
		if(value[0] == 0x00)
		{
			printf("Error: What do I set the value to in line:\n%s\n",line);
            terminate(5);
		}

		bFoundIt = FALSE;

		for(cCountOption = 0; cCountOption < optionlistsize; cCountOption++)
		{
			if(!stricmp(option, optionlist[cCountOption].longname))
			{
				if(optionlist[cCountOption].bSet)
					break;
				int err,temp;
				switch(optionlist[cCountOption].datatype)
				{
				case DATA_NUMBER:
					bFoundIt = TRUE;
					temp = atoi2(value, &err);
					if(!err)
					{
						*(int *)optionlist[cCountOption].ptr = temp;
					}
					else
					{
						fprintf(stderr,"Error non-number input with INI line:\n%s\n", line);
                        terminate(1);
					}
					break;
				case DATA_STRING:
					bFoundIt = TRUE;
					strcpy((char*)optionlist[cCountOption].ptr, value);
					break;
				case DATA_BOOLTRUE:
				case DATA_BOOLFALSE:
					bFoundIt = TRUE;
					if(!stricmp(value,"1") || !stricmp(value,"TRUE"))
						*(int*)optionlist[cCountOption].ptr = TRUE;
					else if(!stricmp(value,"0") || !stricmp(value,"FALSE"))
						*(int*)optionlist[cCountOption].ptr = FALSE;
					else
					{
						printf("Error in reading BOOL value in INI file line:\n%s\n",line);
                        terminate(5);
					}
					break;
				default:
					fprintf(stderr,"Programming error: Unkown option list data type\n");
                    terminate(1);
					break;
				}
			}
		}
		if(bFoundIt == FALSE)
		{
			fprintf(stderr,"Error Inviald opiton in line\n%s\n",line);
            terminate(5);
		}
	}
	fclose(pConfigFile);
}

void ProcessCommandLine(int argc, char *argv[])
{
	int cCountArg, cCountOption;
	char *opt;
	int bFoundIt;

	for( cCountArg = 1; cCountArg < argc; cCountArg++)
	{
		// Is a option
		if(argv[cCountArg][0] == '-')
		{
			if(argv[cCountArg][1] == '?')
			{
				help();
				continue;
			}
			opt = argv[cCountArg];
			opt++;				// Move past -
			bFoundIt = FALSE;
			for(cCountOption = 0; cCountOption < optionlistsize; cCountOption++)
			{
				if(bFoundIt)
					continue;
				char temp[1024];
				strcpy(temp, opt);
				if(optionlist[cCountOption].datatype != 3 &&
				   optionlist[cCountOption].datatype != 4)
					temp[strlen(optionlist[cCountOption].shortname)] = 0x00;
				if(strlen(temp) == 0)
					continue;
				if(!strcmp(temp, optionlist[cCountOption].shortname))
				{
					switch(optionlist[cCountOption].datatype)
					{
					case DATA_NUMBER:
						bFoundIt = TRUE;
						opt += strlen(optionlist[cCountOption].shortname);
						int err,temp;
						temp = atoi2(opt, &err);
						if(!err)
						{
							*(int *)optionlist[cCountOption].ptr = temp;
							optionlist[cCountOption].bSet = TRUE;
						}
						else
						{
							fprintf(stderr,"Error non-number input with parmeter %s\n", argv[cCountArg]);
                            terminate(1);
						}
						break;
					case DATA_STRING:
						bFoundIt = TRUE;
						opt += strlen(optionlist[cCountOption].shortname);
						if( opt[0] == 0x00 )
						{
							fprintf(stderr,"Error: Did not give string for parmater %s\n", argv[cCountArg]);
                            terminate(1);
						}
						strcpy((char*)optionlist[cCountOption].ptr, opt);
						optionlist[cCountOption].bSet = TRUE;
						break;
					case DATA_BOOLTRUE:
						bFoundIt = TRUE;
						*(int*)optionlist[cCountOption].ptr = TRUE;
						optionlist[cCountOption].bSet = TRUE;
						break;
					case DATA_BOOLFALSE:
						bFoundIt = TRUE;
						*(int*)optionlist[cCountOption].ptr = FALSE;
						optionlist[cCountOption].bSet = TRUE;
						break;
					default:
						fprintf(stderr,"Programming error: Unkown option list data type\n");
                        terminate(1);
						break;
					}
				}
			}
			if(bFoundIt == FALSE)
			{
				fprintf(stderr, "Unkown Option %s\n", argv[cCountArg]);
                terminate(1);
			}
		}
		else
		{
			filelist *tflst;
			tflst = (filelist*)malloc(sizeof(struct tag_filelist));
			tflst->next = NULL;
			tflst->fileptr = argv[cCountArg];
			if(FileList == NULL)
			{
				FileList = tflst;
				LastFileList = FileList;
			}
			else
				LastFileList->next = tflst;
		}
	}
}

void testdel(char *del, ...)
{
	char arg[1024];
	vsprintf(arg, del, (void*)(&del+1));
    if(exist(arg) && !keepinter)
        remove(arg);
}

char *pullout(char *out, char *pulling, ...)
{
	vsprintf(out, pulling, (void*)(&pulling+1));
	return out;
}

void RestoreDirectory(void)
{
    chdir(szOldDir);
}

char *_GetMemory(size_t alloc_size, char *file, int line)
{
	char *ret = (char *) malloc(alloc_size);
	if(ret==NULL)
	{
                printf("%s-%d: Unable to allocate %ld bytes\n", file, line,
                        alloc_size);
                terminate(5);
	}
	return ret;
}

int main(int argc, char *argv[])
{
	filelist *tflst;
    char *exepath;

	if(argc == 1)
	{
        help();
		return 1;
	}

	ProcessCommandLine(argc, argv);
	ProcessINIFile(szIniFile);

	if(nCalcMode == -1 ||
	  (nCalcMode != 86 && nCalcMode !=83))
	{
		fprintf(stderr, "Error: Calculator mode not given\n\n");
		help();
        terminate(1);
	}

	if(!nologo)
		printf("GO8X 1.0 (Beta 1) By Chris Dannemiller (c) Copyright 1999\n\n");

	exepath = setuppath(argv[0]);

    strcpy(szOldDir, exepath);

	// Check all files for existance!
	tflst = FileList;
	if(tflst == NULL)
	{
		fprintf(stderr,"No files given!");
        terminate(1);
	}
	while(tflst)
	{
		if(!exist(tflst->fileptr))
		{
			fprintf(stderr,"File %s not found\n",tflst->fileptr);
            terminate(1);
		}
		tflst = tflst->next;
	}

	char *basename = GetMemory(256);
	char *progname = GetMemory(256);
	char *temp1    = GetMemory(256);
	char *temp2    = GetMemory(256);
	char *arg      = GetMemory(256);

	tflst = FileList;
	while(tflst)
	{
        printf("%s:\n",tflst->fileptr);
		// Step 1: Extract file name to no extension
		fnsplit(tflst->fileptr, NULL, NULL, basename, NULL);

		// Step 2: Preprocess
		strcpy(progname, exepath);
		catpath(progname, szExecutablePath);
		catpath(progname, PREPROCESSOR);
        sprintf(arg, "-I%s -D__TI%d__ %s %s.i",
			szIncludePath, nCalcMode, tflst->fileptr, basename);
		pro_spawn_to_file(NULL, NULL, 0, progname, progname, arg, NULL);

		// Step 3: Compile
		strcpy(progname, exepath);
		catpath(progname, szExecutablePath);
		catpath(progname, COMPILER);
        sprintf(arg, "-mTI%d %s.i", nCalcMode, basename);
		pro_spawn_to_file(NULL, NULL, 0, progname, progname, arg, NULL);
		testdel("%s.i",basename);

		// Step 4: Test for compile only
		if(compileonly)
		{
			tflst = tflst->next;
			continue;
		}

		// Step 5: Optimize if allowed if not assebble without opt
		if(!noopt)
		{
			strcpy(progname, exepath);
			catpath(progname, szExecutablePath);
			catpath(progname, OPTIMIZER);
			strcpy(arg, szIncludePath);
			catpath(arg, "RULES");
			pro_spawn_to_file(pullout(temp1, "%s.asm", basename),
					  pullout(temp2, "%s.opt", basename),
					  INPUT | OUTPUT_SCRATCH,
					  progname, progname,
					  arg, NULL);
			testdel("%s.asm", basename);

			strcpy(progname, exepath);
			catpath(progname, szAssemblyPath);

            chdir(progname);

			catpath(progname, ASSEMBLER);
			if(nCalcMode == 86)
				sprintf(arg, "-t80 -b %s%s.opt %s%s", exepath, basename, exepath, basename);
			else if(nCalcMode == 83)
				sprintf(arg, "-t80 -b %s%s.opt %s%s.bin", exepath, basename, exepath, basename);
			pro_spawn_to_file(NULL, NULL, 0, progname, progname, arg, NULL);

            RestoreDirectory();

			testdel("%s.opt", basename);
            testdel("%s.lst", basename);
		}
		else
		{
			strcpy(progname, exepath);
			catpath(progname, szAssemblyPath);

            chdir(progname);

			catpath(progname, ASSEMBLER);
			if(nCalcMode == 86)
                sprintf(arg, "-t80 -b %s%s.asm %s%s", exepath, basename, exepath, basename);
			else if(nCalcMode == 83)
                sprintf(arg, "-t80 -b %s%s.asm %s%s.bin", exepath, basename, exepath, basename);
			pro_spawn_to_file(NULL, NULL, 0, progname, progname, arg, NULL);

            RestoreDirectory();

			testdel("%s.asm", basename);
            testdel("%s.lst", basename);
		}

		// Step 6: Make a program
		strcpy(progname, exepath);
		catpath(progname, szAssemblyPath);

		if(nCalcMode == 86)
		{
			catpath(progname, OBJ86);
			sprintf(arg,"%s",basename);
			pro_spawn_to_file(NULL, NULL, 0, progname, progname, arg, NULL);
            testdel("%s", basename);
		}
		if(nCalcMode == 83)
		{
			catpath(progname, OBJ83);
			sprintf(arg,"%s", basename);
			pro_spawn_to_file(NULL, NULL, 0, progname, progname, arg, NULL);
            testdel("%s.bin", basename);
		}

		tflst = tflst->next;
	}

	free(basename);
	free(progname);
	free(temp1);
	free(temp2);
	free(arg);

	return 0;
}
